
/////////////////////////////////////////////////////////
//
//  zCountLit.cc
//
//  Implements:
//      - Cudd_zddCountLitDouble
//        Counts the number of literals in a set of clauses
//
//
// by Fadi Aloul, University of Michigan
// (c) Copyrights Fadi Aloul, 2002.
//
/////////////////////////////////////////////////////////

#include    "zLib.h"
#include    "st.h"
#include    "cuddInt.h"


/* Fcns */

static double cuddZddCountLitDoubleStep ARGS((DdManager *zdd, DdNode *P, 
					      st_table *table, 
					      DdNode *base, DdNode *empty));
static enum st_retval st_zdd_count_dbl_free ARGS((char *key, char *value, 
						  char *arg));


///////////////////////////////////////////////////////////////////////
// Count the Number of Literals in Set of clauses 
// represented by the ZBDD
//
///////////////////////////////////////////////////////////////////////

double
Cudd_zddCountLitDouble(
  DdManager * zdd,
  DdNode * P)
{
  st_table	*table;
    double	res;
    DdNode	*base, *empty;

    base  = DD_ONE(zdd);
    empty = DD_ZERO(zdd);

    if(P == base)
      return 0;

    table = st_init_table(st_ptrcmp, st_ptrhash);
    if (table == NULL) return((double)CUDD_OUT_OF_MEM);
    res = cuddZddCountLitDoubleStep(zdd, P, table, base, empty);

    if (res == (double)CUDD_OUT_OF_MEM) {
	zdd->errorCode = CUDD_MEMORY_OUT;
    }
    
    st_foreach(table, st_zdd_count_dbl_free, NIL(char));
    st_free_table(table);

    return(res);

} /* end of Cudd_zddCountLitDouble */


static double
cuddZddCountLitDoubleStep(
  DdManager * zdd,
  DdNode * P,
  st_table * table,
  DdNode * base,
  DdNode * empty)
{
    double	res, resT, resE;
    double	*dummy;

    if (P == empty)
	return((double)0.0);
    if (P == base)
      return((double)1.0);

    if (st_lookup(table, (char *)P, (char **)(&dummy))) {
      res = *dummy;
      return(res);
     }

    resE = cuddZddCountLitDoubleStep(zdd, cuddE(P), table, base, empty);
    resT = cuddZddCountLitDoubleStep(zdd, cuddT(P), table, base, empty);

    res = 0;
    if(resE > 0) 
      res = resE;

    if(cuddT(P) == base) {
      res += 1;
    }
    else if(resT > 0) {
      res += (Cudd_zddCountDouble(zdd, cuddT(P)) + resT); 
    }

    dummy = ALLOC(double, 1);
    if (dummy == NULL) {
      return((double)CUDD_OUT_OF_MEM);
    }
    *dummy = res;
    if (st_insert(table, (char *)P, (char *)dummy) == ST_OUT_OF_MEM) {
      FREE(dummy);
      return((double)CUDD_OUT_OF_MEM);
    }

    return(res);
}


static enum st_retval
st_zdd_count_dbl_free(
  char * key,
  char * value,
  char * arg)
{
    double	*d;

    d = (double *)value;
    FREE(d);
    return(ST_CONTINUE);

}
